import _objectWithoutPropertiesLoose from '@babel/runtime/helpers/esm/objectWithoutPropertiesLoose';
import _extends from '@babel/runtime/helpers/esm/extends';
import * as THREE from 'three';
import { Raycaster, Vector2, Scene, WebGLRenderer, PerspectiveCamera } from 'three';
import React, { useRef, useState, useEffect, useCallback, useContext } from 'react';
import Reconciler from 'react-reconciler';
import omit from 'lodash-es/omit';
import upperFirst from 'lodash-es/upperFirst';
import ResizeObserver from 'resize-observer-polyfill';
import { unstable_now, unstable_scheduleCallback, unstable_cancelCallback } from 'scheduler';

const roots = new Map();
const emptyObject = {};
function applyProps(instance, newProps, oldProps) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  // Filter equals, events and reserved props
  const sameProps = Object.keys(newProps).filter(key => newProps[key] === oldProps[key]);
  const handlers = Object.keys(newProps).filter(key => typeof newProps[key] === 'function' && key.startsWith('on'));
  const filteredProps = omit(newProps, [...sameProps, ...handlers, 'children', 'key', 'ref']);

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach((_ref) => {
      let key = _ref[0],
          value = _ref[1];
      let root = instance;
      let target = root[key];

      if (key.includes('-')) {
        const entries = key.split('-');
        target = entries.reduce((acc, key) => acc[key], instance);

        if (target && !target.set) {
          // The target is atomic, this forces us to switch the root
          const _entries$reverse = entries.reverse(),
                name = _entries$reverse[0],
                reverseEntries = _entries$reverse.slice(1);

          root = reverseEntries.reverse().reduce((acc, key) => acc[key], instance);
          key = name;
        }
      }

      if (target && target.set) {
        if (target.constructor.name === value.constructor.name) {
          target.copy(value);
        } else if (Array.isArray(value)) {
          target.set(...value);
        } else {
          target.set(value);
        }
      } else root[key] = value;
    });

    if (handlers.length) {
      instance.__handlers = handlers.reduce((acc, key) => {
        const name = key.charAt(2).toLowerCase() + key.substr(3);
        return _extends({}, acc, {
          [name]: newProps[key]
        });
      }, {}); // Call the update lifecycle, if present

      if (instance.__handlers.update) instance.__handlers.update(instance);
    }
  }
}

function createInstance(type, _ref2) {
  let _ref2$args = _ref2.args,
      args = _ref2$args === void 0 ? [] : _ref2$args,
      props = _objectWithoutPropertiesLoose(_ref2, ["args"]);

  const name = upperFirst(type);
  const instance = type === 'primitive' ? props.object : new THREE[name](...args);
  instance.__objects = [];
  instance.__type = type;
  applyProps(instance, props, {});
  return instance;
}

function appendChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.add(child);else {
      parentInstance.__objects.push(child);

      if (child.name) {
        child.__parent = parentInstance;
        applyProps(parentInstance, {
          [child.name]: child
        });
      }
    }
  }
}

function removeChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.remove(child);else {
      child.__parent = undefined;
      parentInstance.__objects = parentInstance.__objects.filter(c => c !== child);
    }
  }
}

const Renderer = Reconciler({
  now: unstable_now,
  createInstance,
  removeChild,
  appendChild,
  supportsMutation: true,
  isPrimaryRenderer: false,
  schedulePassiveEffects: unstable_scheduleCallback,
  cancelPassiveEffects: unstable_cancelCallback,
  appendInitialChild: appendChild,
  appendChildToContainer: appendChild,
  removeChildFromContainer: removeChild,

  insertBefore(parentInstance, child, beforeChild) {
    if (child) {
      if (child.isObject3D) {
        child.parent = parentInstance;
        child.dispatchEvent({
          type: 'added'
        }); // TODO: the order is out of whack if data objects are presents, has to be recalculated

        const index = parentInstance.children.indexOf(beforeChild);
        parentInstance.children = [...parentInstance.children.slice(0, index), child, ...parentInstance.children.slice(index)];
      } else {
        child.__parent = parentInstance;
        appendChild(parentInstance, child);
      }
    }
  },

  commitUpdate(instance, updatePayload, type, oldProps, newProps, fiber) {
    if (instance.isObject3D) {
      applyProps(instance, newProps, oldProps);
    } else {
      // This is a data object, let's extract critical information about it
      const parent = instance.__parent;

      const _newProps$args = newProps.args,
            argsNew = _newProps$args === void 0 ? [] : _newProps$args,
            restNew = _objectWithoutPropertiesLoose(newProps, ["args"]);

      const _oldProps$args = oldProps.args,
            argsOld = _oldProps$args === void 0 ? [] : _oldProps$args,
            restOld = _objectWithoutPropertiesLoose(oldProps, ["args"]); // If it has new props or arguments, then it needs to be re-instanciated


      if (argsNew.some((value, index) => value !== argsOld[index])) {
        // First it gets removed from its parent
        removeChild(parent, instance); // Next we create a new instance and append it again

        const newInstance = createInstance(instance.type, newProps);
        appendChild(parent, newInstance); // Switch fiber node

        fiber.alternate.stateNode = newInstance;
        fiber.stateNode = newInstance;
      } else {
        // Otherwise just overwrite props
        applyProps(instance, restNew, restOld);
      }
    }
  },

  getPublicInstance(instance) {
    return instance;
  },

  getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },

  getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },

  createTextInstance() {},

  finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },

  prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },

  shouldDeprioritizeSubtree(type, props) {
    return false;
  },

  prepareForCommit() {},

  resetAfterCommit() {},

  shouldSetTextContent(props) {
    return false;
  }

});
function render(element, container) {
  let root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  const root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, () => roots.delete(container));
}

function useMeasure() {
  const ref = useRef();

  const _useState = useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
        bounds = _useState[0],
        set = _useState[1];

  const _useState2 = useState(() => new ResizeObserver((_ref3) => {
    let entry = _ref3[0];
    return set(entry.contentRect);
  })),
        ro = _useState2[0];

  useEffect(() => {
    if (ref.current) ro.observe(ref.current);
    return () => ro.disconnect();
  }, []);
  return [{
    ref
  }, bounds];
}

const context = React.createContext();
const Canvas = React.memo((_ref4) => {
  let children = _ref4.children,
      style = _ref4.style,
      camera = _ref4.camera,
      renderFn = _ref4.render,
      onCreated = _ref4.onCreated,
      onUpdate = _ref4.onUpdate,
      props = _objectWithoutPropertiesLoose(_ref4, ["children", "style", "camera", "render", "onCreated", "onUpdate"]);

  const canvas = useRef();
  const state = useRef({
    subscribers: [],
    active: true,
    canvas: undefined,
    gl: undefined,
    camera: undefined,
    scene: undefined,
    size: undefined,
    subscribe: fn => {
      state.current.subscribers.push(fn);
      return () => state.current.subscribers = state.current.subscribers.filter(s => s === fn);
    }
  });

  const _useMeasure = useMeasure(),
        bind = _useMeasure[0],
        size = _useMeasure[1];

  state.current.size = size;

  const _useState3 = useState(() => new Raycaster()),
        raycaster = _useState3[0];

  const _useState4 = useState(() => new Vector2()),
        mouse = _useState4[0];

  const _useState5 = useState('default'),
        cursor = _useState5[0],
        setCursor = _useState5[1];

  useEffect(() => {
    state.current.scene = window.scene = new Scene();
    state.current.scene.__objects = [];
    state.current.gl = new WebGLRenderer({
      canvas: canvas.current,
      antialias: true,
      alpha: true
    });
    state.current.gl.setClearAlpha(0);
    state.current.camera = camera && camera.current || new PerspectiveCamera(75, 0, 0.1, 1000);
    state.current.gl.setSize(0, 0, false);
    state.current.camera.position.z = 5;
    state.current.canvas = canvas.current;
    if (onCreated) onCreated(state.current);

    const renderLoop = function renderLoop() {
      if (!state.current.active) return;
      requestAnimationFrame(renderLoop);
      if (onUpdate) onUpdate(state.current);
      state.current.subscribers.forEach(fn => fn(state.current));
      if (renderFn) renderFn(state.current);else state.current.gl.render(state.current.scene, state.current.camera);
    }; // Start render-loop


    requestAnimationFrame(renderLoop); // Clean-up

    return () => {
      state.current.active = false;
      unmountComponentAtNode(state.current.scene);
    };
  }, []);
  useEffect(() => {
    state.current.gl.setSize(state.current.size.width, state.current.size.height, false);
    const aspect = state.current.size.width / state.current.size.height;
    state.current.camera.aspect = aspect;
    state.current.camera.updateProjectionMatrix();
    state.current.camera.radius = (state.current.size.width + state.current.size.height) / 4;
  });
  const intersect = useCallback((event, fn) => {
    mouse.x = event.clientX / state.current.size.width * 2 - 1;
    mouse.y = -(event.clientY / state.current.size.height) * 2 + 1;
    raycaster.setFromCamera(mouse, state.current.camera);
    const intersects = raycaster.intersectObjects(state.current.scene.children, true);

    for (var i = 0; i < intersects.length; i++) {
      if (!intersects[i].object.__handlers) continue;
      fn(intersects[i]);
    }

    return intersects;
  });
  useEffect(() => {
    const hovered = {};

    const handleMove = event => {
      let hover = false;
      let intersects = intersect(event, data => {
        const object = data.object;
        const handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(object => {
        if (!intersects.length || !intersects.find(i => i.object === object)) {
          if (object.__handlers.unhover) object.__handlers.unhover();
          delete hovered[object.uuid];
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    return () => {
      window.removeEventListener('mousemove', handleMove);
    };
  }); // Render v-dom into scene

  useEffect(() => {
    if (state.current.size.width > 0) {
      render(React.createElement(context.Provider, {
        value: _extends({}, state.current),
        children: children
      }), state.current.scene);
    }
  }); // Render the canvas into the dom

  return React.createElement("div", _extends({}, bind, props, {
    onClick: event => {
      const clicked = {};
      intersect(event, data => {
        const object = data.object;
        const handlers = object.__handlers;

        if (handlers.click && !clicked[object.uuid]) {
          clicked[object.uuid] = object;
          handlers.click(data);
        }
      });
    },
    style: {
      cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }
  }), React.createElement("canvas", {
    ref: canvas
  }));
});
function useRender(fn) {
  const _useContext = useContext(context),
        subscribe = _useContext.subscribe;

  useEffect(() => subscribe(fn), []);
}
function useThree(fn) {
  const _useContext2 = useContext(context),
        subscribe = _useContext2.subscribe,
        props = _objectWithoutPropertiesLoose(_useContext2, ["subscribe"]);

  return props;
}

export { applyProps, render, unmountComponentAtNode, context, Canvas, useRender, useThree };
